import sys
from pathlib import Path

TESTS_DIRECTORY = Path(__file__).resolve().parent
PROJECT_ROOT = TESTS_DIRECTORY.parent
TEST_DB_JSON_PATH = TESTS_DIRECTORY / "test-db.json"

sys.path.append(str(PROJECT_ROOT))

from typing import Iterable

import notmongo as nm
from notmongo.typedefs import *

# TODO:
#   - test operators on different datatypes: string < string, id, ...
#   - test field.names.with.dots
#   - read the mongodb operator docs and test edge cases


def _make_collection(documents: Iterable[BsonDoc]) -> nm.SyncCollection:
    db = nm.SyncDatabase()
    col = db.collection("test-advanced-query-collection")
    col.insert_many(documents)
    return col


def _load_test_collection(collection_name: str = "animals") -> nm.SyncCollection:
    db = nm.SyncDatabase.load_json(TEST_DB_JSON_PATH)
    return db.collection(collection_name)


def _verify_result_ids(
    documents: Union[nm.SyncCollection, Iterable[BsonLike]],
    expected_ids: Iterable[BsonLike],
):
    # Get all ids
    if isinstance(documents, nm.SyncCollection):
        documents = documents.find({})

    have_ids = {doc["_id"] for doc in documents}

    expected_ids = set(expected_ids)

    # Superfluous
    superfluous_ids = have_ids - expected_ids
    assert (
        not superfluous_ids
    ), f"Superfluous result items, with these ids: `{'`, `'.join(map(str, superfluous_ids))}`"

    # Missing
    missing_ids = expected_ids - have_ids
    assert (
        not missing_ids
    ), f"Missing result items, with these ids: `{'`, `'.join(map(str, missing_ids))}`"


def _verify_comparison_operator(
    query: Any, values_in_result: Iterable[Any], values_not_in_result: Iterable[Any]
):
    values_in_result = list(values_in_result)
    values_not_in_result = list(values_not_in_result)

    # Insert the values into a new collection
    col = _make_collection(
        [{"testField": val} for val in values_in_result + values_not_in_result]
    )

    # Apply the queries
    result = col.find(query)
    result = list(result)

    # Compare
    values_returned = [val["testField"] for val in result]
    values_returned.sort()
    values_in_result.sort()
    assert values_returned == values_in_result


def test_operator_and_empty():
    col = _load_test_collection()
    result = col.find({"$and": []})
    _verify_result_ids(result, {"id-1-cat", "id-1-dog", "id-1-fish"})


def test_operator_and():
    col = _load_test_collection()
    result = col.find(
        {
            "$and": [
                {"canSwim": True},
                {"age": {"$gt": 2}},
            ]
        }
    )
    _verify_result_ids(result, {"id-1-dog"})


def test_operator_or():
    col = _load_test_collection()
    result = col.find(
        {
            "$or": [
                {"_id": "id-1-cat"},
                {"_id": "id-1-dog"},
            ]
        }
    )
    _verify_result_ids(result, {"id-1-cat", "id-1-dog"})


def test_operator_or_empty():
    col = _load_test_collection()
    result = col.find({"$or": []})
    _verify_result_ids(result, {})


def test_operator_lt_numeric():
    _verify_comparison_operator(
        {"testField": {"$lt": 3}},
        [1, 2],
        [3, 4, 5],
    )


def test_operator_lte_numeric():
    _verify_comparison_operator(
        {"testField": {"$lte": 3}},
        [1, 2, 3],
        [4, 5],
    )


def test_operator_eq_numeric():
    _verify_comparison_operator(
        {"testField": {"$eq": 3}},
        [3],
        [1, 2, 4, 5],
    )


def test_operator_ne_numeric():
    _verify_comparison_operator(
        {"testField": {"$ne": 3}},
        [1, 2, 4, 5],
        [3],
    )


def test_operator_gte_numeric():
    _verify_comparison_operator(
        {"testField": {"$gte": 3}},
        [3, 4, 5],
        [1, 2],
    )


def test_operator_gt_numeric():
    _verify_comparison_operator(
        {"testField": {"$gt": 3}},
        [4, 5],
        [1, 2, 3],
    )


def test_operator_not():
    col = _load_test_collection()
    result = col.find({"$not": {"_id": "id-1-cat"}})
    _verify_result_ids(result, {"id-1-dog", "id-1-fish"})


def test_operator_exists_positive():
    col = _load_test_collection()
    result = col.find({"onlyLandwalkerField": {"$exists": True}})
    _verify_result_ids(result, {"id-1-cat", "id-1-dog"})


def test_operator_exists_negative():
    col = _load_test_collection()
    result = col.find({"onlyLandwalkerField": {"$exists": False}})
    _verify_result_ids(result, {"id-1-fish"})


def test_operator_implicit_eq_scalar_with_array():
    col = _load_test_collection("homes")
    result = col.find({"entrances": "front"})
    _verify_result_ids(result, {"id-2-doghouse"})


def test_operator_explicit_eq_scalar_with_array():
    col = _load_test_collection("homes")
    result = col.find({"entrances": {"$eq": "front"}})
    _verify_result_ids(result, {"id-2-doghouse"})


def test_operator_ne_scalar_with_array():
    col = _load_test_collection("homes")
    result = col.find({"entrances": {"$ne": "front"}})
    _verify_result_ids(result, {"id-2-aquarium"})



import tempfile
from pathlib import Path

import pytest

import notmongo as nm
from notmongo.typedefs import *

TESTS_DIRECTORY = Path(__file__).resolve().parent
TEST_DB_JSON_PATH = TESTS_DIRECTORY / "test-db.json"


def _load_test_db() -> nm.AsyncDatabase:
    return nm.AsyncDatabase.load_json(TEST_DB_JSON_PATH)


@pytest.mark.asyncio
async def test_file_persistence():
    # Create a DB
    db = _load_test_db()

    # Set up the persistence
    with tempfile.NamedTemporaryFile() as temp_file:
        file_path = Path(temp_file.name)

    assert file_path.parent.exists()
    assert not file_path.exists()

    db.persistence = nm.FilePersistence(
        file_path,
        n_backups=2,
        file_format="json",
        save_interval=1000,
        format_json=False,
    )

    # There shouldn't be a file yet
    assert (
        not file_path.exists()
    ), "The file has been created before any changes to the DB were made"

    # Perform any mutation in the db
    await db["test-collection"].insert_one({"_id": "some-id"})

    # Now there has to be a file
    assert file_path.exists(), "No file was written, despite mutating changes to the DB"



import ctypes as ct
import importlib.util
import os
import platform
from pathlib import Path

import bson

from . import meta
from .typedefs import *


class wrapped_bson(ct.Structure):
    _fields_ = [
        ("value", ct.c_void_p),
        ("value_length", ct.c_size_t),
    ]

    @staticmethod
    def from_bytes(bson_bytes: bytes) -> "wrapped_bson":
        # Encode the value to bytes
        passed_buffer = malloc(len(bson_bytes))
        ct.memmove(passed_buffer, bson_bytes, len(bson_bytes))

        # Construct the result
        result = wrapped_bson()
        result.value = passed_buffer
        result.value_length = len(bson_bytes)

        return result

    @classmethod
    def from_bson(cls, bson_obj: BsonLike) -> "wrapped_bson":
        # Encode the value to bytes
        return cls.from_bytes(bson.encode(bson_obj))  # type: ignore

    def consume_into_bson(self) -> BsonDoc:
        # Get the pointed-to bytes as bytes object
        unowned_bytes = (ct.c_char * self.value_length).from_address(self.value)
        owned_bytes = bytes(unowned_bytes)

        # Parse the BSON
        return bson.decode(owned_bytes)

    def consume_into_response(self) -> BsonLike:
        result = self.consume_into_bson()
        if result["success"]:
            return result["result"]
        else:
            raise NotmongoError(result["errorMessage"])  # type: ignore


def compression_to_int(compression: Compression) -> int:
    if compression is None:
        return 0

    if compression == "zip":
        return 1

    raise ValueError(f"Unknown compression type: {compression}")


# Load the native library
def _load_so_file() -> ct.CDLL:
    # If the environment variable "NOTMONGO_SO_PATH" is set, use that
    if "NOTMONGO_SO_PATH" in os.environ:
        so_path = Path(os.environ["NOTMONGO_SO_PATH"])

        if not so_path.exists():
            raise FileNotFoundError(
                f"The path to the native library was explicitly set using the environment variable `NOTMONGO_SO_PATH`, but the file does not exist: {so_path}"
            )

    # Otherwise fetch it from the package's shared_objects directory
    else:
        SHARED_OBJECTS_DIR = Path(__file__).parent / "shared_objects"
        so_path = SHARED_OBJECTS_DIR / "libnotmongo.so"

        if not so_path.exists():
            raise FileNotFoundError(
                f"The native library was not found at the expected path: {so_path}"
            )


    return ct.CDLL(str(so_path))


SO_FILE = _load_so_file()


# Helper functions

test_function = SO_FILE.notmongo_test_function
test_function.argtypes = [ct.c_size_t]
test_function.restype = ct.c_size_t

assert test_function(1) == 2, "Test function failed"

notmongo_get_metadata = SO_FILE.notmongo_get_metadata
notmongo_get_metadata.argtypes = []
notmongo_get_metadata.restype = wrapped_bson

SO_META = notmongo_get_metadata().consume_into_bson()

_so_version = SO_META["version"]
_py_version = meta.__version__
assert (
    _so_version == _py_version
), f"The native library version does not match that of the python library: {_so_version} vs {_py_version}"

malloc = SO_FILE.notmongo_malloc
malloc.argtypes = [ct.c_size_t]
malloc.restype = ct.c_void_p

free = SO_FILE.notmongo_free
free.argtypes = [ct.c_void_p]
free.restype = None

# Database creation / destruction

db_new_empty = SO_FILE.notmongo_db_new_empty
db_new_empty.argtypes = []
db_new_empty.restype = ct.c_void_p

db_free = SO_FILE.notmongo_db_free
db_free.argtypes = [ct.c_void_p]
db_free.restype = None

# Data loading

db_populate_from_bson_value = SO_FILE.notmongo_db_populate_from_bson_value
db_populate_from_bson_value.argtypes = [ct.c_void_p, wrapped_bson]
db_populate_from_bson_value.restype = wrapped_bson

db_populate_from_bson_file = SO_FILE.notmongo_db_populate_from_bson_file
db_populate_from_bson_file.argtypes = [ct.c_void_p, ct.c_char_p, ct.c_uint32]
db_populate_from_bson_file.restype = wrapped_bson

db_populate_from_json_string = SO_FILE.notmongo_db_populate_from_json_string
db_populate_from_json_string.argtypes = [ct.c_void_p, wrapped_bson]
db_populate_from_json_string.restype = wrapped_bson

db_populate_from_json_file = SO_FILE.notmongo_db_populate_from_json_file
db_populate_from_json_file.argtypes = [ct.c_void_p, ct.c_char_p, ct.c_uint32]
db_populate_from_json_file.restype = wrapped_bson

# Data dumping

db_dump_to_json_file = SO_FILE.notmongo_db_dump_to_json_file
db_dump_to_json_file.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    ct.c_bool,
    ct.c_bool,
    ct.c_uint32,
    ct.c_uint32,
]
db_dump_to_json_file.restype = wrapped_bson

db_dump_to_json_string = SO_FILE.notmongo_db_dump_to_json_string
db_dump_to_json_string.argtypes = [ct.c_void_p, ct.c_bool, ct.c_bool]
db_dump_to_json_string.restype = wrapped_bson

db_dump_to_bson_file = SO_FILE.notmongo_db_dump_to_bson_file
db_dump_to_bson_file.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    ct.c_uint32,
    ct.c_uint32,
]
db_dump_to_bson_file.restype = wrapped_bson

db_dump_to_bson_value = SO_FILE.notmongo_db_dump_to_bson_value
db_dump_to_bson_value.argtypes = [ct.c_void_p]
db_dump_to_bson_value.restype = wrapped_bson

# Query: Database Meta

db_drop_collection = SO_FILE.notmongo_db_drop_collection
db_drop_collection.argtypes = [ct.c_void_p, ct.c_char_p]
db_drop_collection.restype = wrapped_bson

db_n_collections = SO_FILE.notmongo_db_n_collections
db_n_collections.argtypes = [ct.c_void_p]
db_n_collections.restype = wrapped_bson

db_collection_names = SO_FILE.notmongo_db_collection_names
db_collection_names.argtypes = [ct.c_void_p]
db_collection_names.restype = wrapped_bson

# Query: Collection Meta

collection_count_all_documents = SO_FILE.notmongo_collection_count_all_documents
collection_count_all_documents.argtypes = [ct.c_void_p, ct.c_char_p]
collection_count_all_documents.restype = wrapped_bson

collection_count_documents = SO_FILE.notmongo_collection_count_documents
collection_count_documents.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
    ct.c_size_t,
    ct.c_size_t,
]
collection_count_documents.restype = wrapped_bson

# Query: Delete

collection_delete_one = SO_FILE.notmongo_collection_delete_one
collection_delete_one.argtypes = [ct.c_void_p, ct.c_char_p, wrapped_bson]
collection_delete_one.restype = wrapped_bson

collection_delete_many = SO_FILE.notmongo_collection_delete_many
collection_delete_many.argtypes = [ct.c_void_p, ct.c_char_p, wrapped_bson]
collection_delete_many.restype = wrapped_bson

# Query: Find

collection_find = SO_FILE.notmongo_collection_find
collection_find.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
    wrapped_bson,
    ct.c_size_t,
    ct.c_size_t,
    wrapped_bson,
]
collection_find.restype = wrapped_bson

collection_find_one = SO_FILE.notmongo_collection_find_one
collection_find_one.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
    wrapped_bson,
    ct.c_size_t,
    wrapped_bson,
]
collection_find_one.restype = wrapped_bson

# Query: Insert

collection_insert_one = SO_FILE.notmongo_collection_insert_one
collection_insert_one.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
]
collection_insert_one.restype = wrapped_bson

collection_insert_many = SO_FILE.notmongo_collection_insert_many
collection_insert_many.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
]
collection_insert_many.restype = wrapped_bson

# Query: Replace

collection_replace_one = SO_FILE.notmongo_collection_replace_one
collection_replace_one.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
    wrapped_bson,
    ct.c_bool,
]
collection_replace_one.restype = wrapped_bson

# Query: Update

collection_update_one = SO_FILE.notmongo_collection_update_one
collection_update_one.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
    wrapped_bson,
    ct.c_bool,
]
collection_update_one.restype = wrapped_bson

collection_update_many = SO_FILE.notmongo_collection_update_many
collection_update_many.argtypes = [
    ct.c_void_p,
    ct.c_char_p,
    wrapped_bson,
    wrapped_bson,
    ct.c_bool,
]
collection_update_many.restype = wrapped_bson



import json
import random
import tempfile
import zipfile
from pathlib import Path
from typing import *  # type: ignore

import bson
import bson.errors
import pytest

import notmongo as nm

TESTS_DIRECTORY = Path(__file__).resolve().parent
TEST_DB_JSON_PATH = TESTS_DIRECTORY / "test-db.json"

COMPRESSIONS = [None, "zip"]


def _load_test_db_dict() -> Dict[str, Any]:
    with open(TEST_DB_JSON_PATH) as f:
        return json.load(f)


async def _verify_loaded_db(db: nm.AsyncDatabase):
    n_documents = {"animals": 3, "homes": 2}

    # Correct number of collections
    assert await db.n_collections() == len(n_documents)

    # Collections have the correct names and number of documents
    async for collection in db.collections():
        assert await collection.count_all_documents() == n_documents[collection.name]


def _as_file(
    contents: Union[bytes, str],
    compression: nm.Compression,
) -> Path:
    # Make sure the contents are bytes
    if isinstance(contents, str):
        contents = contents.encode("utf-8")

    if not isinstance(contents, bytes):
        raise TypeError(contents)

    # Write to a temporary file
    if compression is None:
        f = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        f.write(contents)
        return Path(f.name)

    if compression == "zip":
        # Write to a temporary file
        f = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        with zipfile.ZipFile(f, "w") as zip_file:
            zip_file.writestr(f"some_random_name_{random.random()}.bson", contents)

        return Path(f.name)

    assert False, f"Invalid compression: `{compression}`"


def _read_compressed(path: Path, compression: nm.Compression) -> bytes:
    if compression is None:
        return path.read_bytes()

    if compression == "zip":
        with zipfile.ZipFile(path, "r") as zip_file:
            assert len(zip_file.namelist()) == 1
            return zip_file.read(zip_file.namelist()[0])

    assert False, f"Invalid compression: `{compression}`"


@pytest.mark.asyncio
async def test_handling_db():
    db = nm.AsyncDatabase()
    collection = db.collection("test-collection")
    await collection.insert_one({"key": "value"})
    assert await db.n_collections() == 1

    async for also_collection in db.collections():
        break

    assert collection.name == also_collection.name  # type: ignore

    await collection.drop()
    assert await db.n_collections() == 0


@pytest.mark.asyncio
async def test_loads_valid_bson():
    as_dict = _load_test_db_dict()
    as_bson = bson.encode(as_dict)
    db = nm.AsyncDatabase.loads_bson(as_bson)

    await _verify_loaded_db(db)


@pytest.mark.asyncio
async def test_loads_invalid_bson():
    as_bson = b"1234567890"

    with pytest.raises((nm.NotmongoError, bson.errors.InvalidBSON)):
        nm.AsyncDatabase.loads_bson(as_bson)


@pytest.mark.asyncio
@pytest.mark.parametrize("compression", COMPRESSIONS)
async def test_load_valid_bson(compression: nm.Compression):
    as_dict = _load_test_db_dict()
    as_bson = bson.encode(as_dict)
    as_path = _as_file(as_bson, compression=compression)

    db = nm.AsyncDatabase.load_bson(as_path, compression=compression)

    await _verify_loaded_db(db)


@pytest.mark.asyncio
@pytest.mark.parametrize("compression", COMPRESSIONS)
async def test_load_invalid_bson(compression: nm.Compression):
    as_bson = b"1234567890"
    as_path = _as_file(as_bson, compression=compression)

    with pytest.raises((nm.NotmongoError, bson.errors.InvalidBSON)):
        nm.AsyncDatabase.load_bson(as_path, compression=compression)


@pytest.mark.asyncio
async def test_loads_valid_json():
    as_dict = _load_test_db_dict()
    as_json = json.dumps(as_dict)
    db = nm.AsyncDatabase.loads_json(as_json)

    await _verify_loaded_db(db)


@pytest.mark.asyncio
async def test_loads_invalid_json_string():
    as_json = "{foo?"

    with pytest.raises(nm.NotmongoError):
        nm.AsyncDatabase.loads_json(as_json)


@pytest.mark.asyncio
@pytest.mark.parametrize("compression", COMPRESSIONS)
async def test_load_valid_json(compression: nm.Compression):
    as_dict = _load_test_db_dict()
    as_json = json.dumps(as_dict)
    as_path = _as_file(as_json, compression)

    db = nm.AsyncDatabase.load_json(as_path, compression=compression)

    await _verify_loaded_db(db)


@pytest.mark.asyncio
@pytest.mark.parametrize("compression", COMPRESSIONS)
async def test_load_invalid_json(compression: nm.Compression):
    as_path = _as_file("{foo?", compression)

    with pytest.raises(nm.NotmongoError):
        nm.AsyncDatabase.load_json(as_path, compression=compression)


@pytest.mark.asyncio
async def test_dumps_bson():
    as_dict_orig = _load_test_db_dict()
    as_bson_orig = bson.encode(as_dict_orig)
    db = nm.AsyncDatabase.loads_bson(as_bson_orig)

    as_bson_dump = db.dumps_bson()
    as_dict_dump = bson.decode(as_bson_dump)

    assert as_dict_orig == as_dict_dump


@pytest.mark.asyncio
@pytest.mark.parametrize("compression", COMPRESSIONS)
async def test_dump_bson(compression: nm.Compression):
    as_dict_orig = _load_test_db_dict()
    as_bson_orig = bson.encode(as_dict_orig)
    db = nm.AsyncDatabase.loads_bson(as_bson_orig)

    with tempfile.NamedTemporaryFile() as f:
        temp_path = Path(f.name)

    db.dump_bson(temp_path, compression=compression)
    as_dict_dump = bson.decode(_read_compressed(temp_path, compression))

    assert as_dict_orig == as_dict_dump


@pytest.mark.asyncio
async def test_dumps_json():
    as_dict_orig = _load_test_db_dict()
    as_bson_orig = bson.encode(as_dict_orig)
    db = nm.AsyncDatabase.loads_bson(as_bson_orig)

    as_json_dump = db.dumps_json()
    as_dict_dump = json.loads(as_json_dump)

    assert as_dict_orig == as_dict_dump


@pytest.mark.asyncio
@pytest.mark.parametrize("compression", COMPRESSIONS)
async def test_dump_json(compression: nm.Compression):
    as_dict_orig = _load_test_db_dict()
    as_bson_orig = bson.encode(as_dict_orig)
    db = nm.AsyncDatabase.loads_bson(as_bson_orig)

    with tempfile.NamedTemporaryFile() as f:
        temp_path = Path(f.name)

    db.dump_json(temp_path, compression=compression)
    as_dict_dump = json.loads(_read_compressed(temp_path, compression).decode("utf-8"))

    assert as_dict_orig == as_dict_dump



from typing import *  # type: ignore

BsonLike: TypeAlias = Union[
    Dict[str, "BsonLike"], List["BsonLike"], str, int, float, bool, None
]

BsonDoc: TypeAlias = Dict[str, BsonLike]

SortList: TypeAlias = List[Tuple[str, Literal[-1, 1]]]

Compression: TypeAlias = Union[None, Literal["zip"]]


class NotmongoError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __repr__(self):
        return f"<NotMongoException {repr(self.message)}>"



import random
from pathlib import Path
from typing import Iterable

import notmongo as nm
from notmongo.typedefs import *

TESTS_DIRECTORY = Path(__file__).resolve().parent
TEST_DB_JSON_PATH = TESTS_DIRECTORY / "test-db.json"


def _verify_sorting(documents: Iterable[BsonLike], sort: SortList):
    """
    Makes sure that documents are sorted in the expected order. The documents
    are inserted into an empty collection, and queried with the given sort. If
    any documents are not returned in the same order as they were passed to this
    function an AssertionError is raised.
    """

    # Randomize the document order, so they aren't sorted correctly merely by
    # accident. Make sure to do this deterministically so the test results are
    # reproducible.
    documents = list(documents)
    shuffled_documents = documents.copy()
    random.Random(0).shuffle(shuffled_documents)

    # If shuffling randomly led to a correct order, reverse their order
    if documents == shuffled_documents:
        shuffled_documents.reverse()

    # Insert the documents into an empty database
    db = nm.SyncDatabase()
    col = db.collection("sorting-test")
    col.insert_many(shuffled_documents)

    # Query the documents with the given sort
    query_result = col.find(sort=sort, projection={"_id": False})
    query_result = list(query_result)

    # assert id_order_should == id_order_got
    assert documents == query_result


def test_single_field_name():
    _verify_sorting(
        [
            {"value": "bird"},
            {"value": "cat"},
            {"value": "dog"},
            {"value": "eel"},
        ],
        "value",
    )


def test_string():
    _verify_sorting(
        [
            {"value": "bird"},
            {"value": "cat"},
            {"value": "dog"},
            {"value": "eel"},
        ],
        [("value", 1)],
    )


def test_numeric_descending():
    _verify_sorting(
        [
            {"value": 4},
            {"value": 3},
            {"value": 2},
            {"value": 1},
        ],
        [("value", -1)],
    )


def test_different_numeric_types():
    _verify_sorting(
        [
            {"value": -2},
            {"value": -1.5},
            {"value": -1},
            {"value": False},
            {"value": True},
            {"value": 2},
            {"value": 2.5},
            {"value": 2**40},  # Doesn't fit into a 32-Bit Integer
            {"value": 2**60},  # Oudside of a Double's continuous integer range
        ],
        [("value", 1)],
    )


def test_multiple():
    _verify_sorting(
        [
            {"b1": 0, "b2": 0, "b3": 0},
            {"b1": 0, "b2": 0, "b3": 1},
            {"b1": 0, "b2": 1, "b3": 0},
            {"b1": 0, "b2": 1, "b3": 1},
            {"b1": 1, "b2": 0, "b3": 0},
            {"b1": 1, "b2": 0, "b3": 1},
            {"b1": 1, "b2": 1, "b3": 0},
            {"b1": 1, "b2": 1, "b3": 1},
        ],
        [("b1", 1), ("b2", 1), ("b3", 1)],
    )


# TODO:
# - Other sortables:
#   - DateTime
#   - Binary (important for uuids!)
#   - ObjectId
#   - Boolean
# - Special cases
#   - arrays
# - connect to actual mongodb and compare the orders? Maybe even fuzz?



from pathlib import Path
from typing import Iterable

import pytest

import notmongo as nm
from notmongo.typedefs import *

TESTS_DIRECTORY = Path(__file__).resolve().parent
TEST_DB_JSON_PATH = TESTS_DIRECTORY / "test-db.json"


def _verify_projection(document: BsonLike, projection, expected_fields: Iterable[str]):
    """
    Verifies that projections work as expected, by inserting a document into a
    empy database, querying it, and making sure exactly the correct fields are
    present.
    """

    # Insert the documents int an empty database
    db = nm.SyncDatabase()
    col = db.collection("projection-test")
    col.insert_one(document)

    # Query the document with the given projection
    query_result = col.find(projection=projection)
    query_result = list(query_result)
    assert len(query_result) == 1
    query_result = query_result[0]

    # Prepare some values for easy checking
    expected_fields = set(expected_fields)
    have_fields = set(query_result.keys())

    # Superfluous
    superfluous_fields = have_fields - expected_fields
    assert (
        not superfluous_fields
    ), f"Superfluous fields in the result: `{'`, `'.join(map(str, superfluous_fields))}`"

    # Missing
    missing_fields = expected_fields - have_fields
    assert (
        not missing_fields
    ), f"Missing fields in the result: `{'`, `'.join(map(str, missing_fields))}`"


def test_keep_all_fields_implicit():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {},
        {"_id", "a", "b"},
    )


def test_keep_all_fields_explicit():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {"_id": True, "a": True, "b": True},
        {"_id", "a", "b"},
    )


def test_drop_all_fields():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {"_id": False, "a": False, "b": False},
        {},
    )


def test_keep_some_rest_implicit():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {"_id": True, "a": True},
        {"_id", "a"},
    )


def test_drop_some_rest_implicit():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {"_id": False, "a": False},
        {"b"},
    )


def test_keep_some_implicit_id():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {"a": True},
        {"_id", "a"},
    )


def test_drop_some_implicit_id():
    _verify_projection(
        {"_id": None, "a": None, "b": None},
        {"a": False},
        {"_id", "b"},
    )


def test_invalid_projection():
    with pytest.raises(nm.NotmongoError):
        _verify_projection(
            {"_id": None, "a": None, "b": None},
            {"a": False, "b": True},  # Some fields are true, some false
            {"_id", "b"},
        )



import asyncio
from pathlib import Path
from typing import AsyncGenerator, Optional

from . import async_collection, persistence, sync_database
from .typedefs import *


class AsyncDatabase:
    """
    Asynchronous equivalent for SyncDatabase. Method calls are  the same as for
    SyncDatabase , but instead of SyncCollections, this class returns
    AsyncCollections.


    This doesn't provide the the advantages of actual async operations, but
    allows code to be written asynchronously, and then easily switched to the
    real MongoDB.
    """

    def __init__(self):
        self._sync_db = sync_database.SyncDatabase()
        self._db_lock = asyncio.Lock()

    @property
    def persistence(self) -> Optional[persistence.BasePersistence]:
        return self._sync_db.persistence

    @persistence.setter
    def persistence(self, value: "persistence.BasePersistence"):
        self._sync_db.persistence = value

    @classmethod
    def load_json(
        cls,
        path: Union[Path, str],
        *,
        compression: Compression = None,
    ) -> "AsyncDatabase":
        # TODO: Make async?
        result = AsyncDatabase()
        result._sync_db = sync_database.SyncDatabase.load_json(
            path,
            compression=compression,
        )
        return result

    @classmethod
    def loads_json(cls, serialized: str) -> "AsyncDatabase":
        # TODO: Make async?
        result = AsyncDatabase()
        result._sync_db = sync_database.SyncDatabase.loads_json(serialized)
        return result

    @classmethod
    def load_bson(
        cls,
        path: Path,
        compression: Compression,
    ) -> "AsyncDatabase":
        # TODO: Make async?
        result = AsyncDatabase()
        result._sync_db = sync_database.SyncDatabase.load_bson(
            path,
            compression=compression,
        )
        return result

    @classmethod
    def loads_bson(cls, contents: bytes) -> "AsyncDatabase":
        # TODO: Make async?
        result = AsyncDatabase()
        result._sync_db = sync_database.SyncDatabase.loads_bson(contents)
        return result

    def dump_json(
        self,
        path: Path,
        *,
        relaxed: bool = True,
        format: bool = False,
        n_backups: int = 0,
        compression: Compression = None,
    ) -> None:
        # TODO: Make async?
        self._sync_db.dump_json(
            path,
            relaxed=relaxed,
            format=format,
            n_backups=n_backups,
            compression=compression,
        )

    def dumps_json(self, relaxed: bool = True, format: bool = False) -> str:
        # TODO: Make async?
        return self._sync_db.dumps_json(relaxed=relaxed, format=format)

    def dump_bson(
        self,
        path: Path,
        *,
        n_backups: int = 0,
        compression: Compression = None,
    ) -> None:
        # TODO: Make async?
        self._sync_db.dump_bson(
            path,
            n_backups=n_backups,
            compression=compression,
        )

    def dumps_bson(self) -> bytes:
        # TODO: Make async?
        return self._sync_db.dumps_bson()

    async def n_collections(self) -> int:
        # This is a fast operation, no point in using a thread
        return self._sync_db.n_collections()

    async def collections(
        self,
    ) -> AsyncGenerator["async_collection.AsyncCollection", None]:
        # This is a fast operation, no point in using a thread
        for scol in self._sync_db.collections():
            yield async_collection.AsyncCollection(scol, self._db_lock)

    def collection(self, collection_name: str) -> "async_collection.AsyncCollection":
        if not isinstance(collection_name, str):
            raise TypeError("The collection name has to be a string.")

        return async_collection.AsyncCollection(
            self._sync_db.collection(collection_name),
            self._db_lock,
        )

    async def drop_collection(self, collection_name: str) -> None:
        # This is a fast operation, no point in using a thread
        self._sync_db.drop_collection(collection_name)

    def __getitem__(self, collection_name: str) -> "async_collection.AsyncCollection":
        return self.collection(collection_name)



import ctypes
from pathlib import Path
from typing import *  # type: ignore

import bson

from . import native, persistence, sync_collection
from .typedefs import *


class SyncDatabase:
    def __init__(self):
        self._rust_db: ctypes.c_void_p = native.db_new_empty()
        self._persistence: Optional[persistence.BasePersistence] = None

    @property
    def persistence(self) -> Optional[persistence.BasePersistence]:
        return self._persistence

    @persistence.setter
    def persistence(self, value: "persistence.BasePersistence"):
        if not isinstance(value, persistence.BasePersistence):
            raise TypeError(value)

        if self._persistence is not None:
            self._persistence.save_now()
            self._persistence.db = None

        self._persistence = value
        self._persistence.db = self

    @classmethod
    def load_json(
        cls,
        path: Union[Path, str],
        *,
        compression: Compression = None,
    ) -> "SyncDatabase":
        # FIXME: "load" should really accept a file-like object, not a path!

        if isinstance(path, str):
            path = Path(path)

        if not isinstance(path, Path):
            raise TypeError("The path has to be a Path or a string.")

        # Load the contents
        result = SyncDatabase()
        call_result = native.db_populate_from_json_file(
            result._rust_db,
            str(path.resolve()).encode("utf-8"),
            native.compression_to_int(compression),
        )

        # Check for errors
        call_result.consume_into_response()
        return result

    @classmethod
    def loads_json(cls, serialized: str) -> "SyncDatabase":
        if not isinstance(serialized, str):
            raise TypeError(serialized)

        # Load the contents
        result = SyncDatabase()
        call_result = native.db_populate_from_json_string(
            result._rust_db,
            native.wrapped_bson.from_bson({"contents": serialized}),
        )

        # Check for errors
        call_result.consume_into_response()
        return result

    @classmethod
    def load_bson(
        cls,
        path: Path,
        *,
        compression: Compression,
    ) -> "SyncDatabase":
        # FIXME: "load" should really accept a file-like object, not a path!

        if isinstance(path, str):
            path = Path(path)

        if not isinstance(path, Path):
            raise TypeError("The path has to be a Path or a string.")

        # Load the contents
        result = SyncDatabase()
        call_result = native.db_populate_from_bson_file(
            result._rust_db,
            str(path.resolve()).encode("utf-8"),
            native.compression_to_int(compression),
        )

        # Check for errors
        call_result.consume_into_response()
        return result

    @classmethod
    def loads_bson(cls, contents: bytes) -> "SyncDatabase":
        # Load the contents
        result = SyncDatabase()
        call_result = native.db_populate_from_bson_value(
            result._rust_db, native.wrapped_bson.from_bson(bson.decode(contents))
        )

        # Check for errors
        call_result.consume_into_response()
        return result

    def dump_json(
        self,
        path: Path,
        *,
        relaxed: bool = True,
        format: bool = False,
        n_backups: int = 0,
        compression: Compression = None,
    ) -> None:
        # FIXME: "dump" should really accept a file-like object, not a path!

        if isinstance(path, str):
            path = Path(path)

        if not isinstance(path, Path):
            raise TypeError("The path has to be a Path or a string.")

        if not isinstance(relaxed, bool):
            raise TypeError(relaxed)

        if not isinstance(format, bool):
            raise TypeError(format)

        if not isinstance(n_backups, int):
            raise TypeError(n_backups)

        if n_backups < 0:
            raise ValueError(n_backups)

        call_result = native.db_dump_to_json_file(
            self._rust_db,
            str(path.resolve()).encode("utf-8"),
            relaxed,
            format,
            n_backups,
            native.compression_to_int(compression),
        )
        call_result.consume_into_response()

    def dumps_json(
        self,
        *,
        relaxed: bool,
        format: bool = False,
    ) -> str:
        if not isinstance(relaxed, bool):
            raise TypeError(relaxed)

        if not isinstance(format, bool):
            raise TypeError(format)

        call_result = native.db_dump_to_json_string(self._rust_db, relaxed, format)
        return call_result.consume_into_response()

    def dump_bson(
        self,
        path: Path,
        *,
        n_backups: int = 0,
        compression: Compression = None,
    ) -> None:
        # FIXME: "dump" should really accept a file-like object, not a path!

        if isinstance(path, str):
            path = Path(path)

        if not isinstance(path, Path):
            raise TypeError("The path has to be a Path or a string.")

        if not isinstance(n_backups, int):
            raise TypeError(n_backups)

        if n_backups < 0:
            raise ValueError(n_backups)

        call_result = native.db_dump_to_bson_file(
            self._rust_db,
            str(path.resolve()).encode("utf-8"),
            n_backups,
            native.compression_to_int(compression),
        )
        call_result.consume_into_response()

    def dumps_bson(self) -> bytes:
        call_result = native.db_dump_to_bson_value(
            self._rust_db,
        )
        return bson.encode(call_result.consume_into_response())

    def n_collections(self) -> int:
        call_result = native.db_n_collections(self._rust_db)
        return call_result.consume_into_response()

    def collections(self) -> Iterable["sync_collection.SyncCollection"]:
        call_result = native.db_collection_names(self._rust_db)
        call_result = call_result.consume_into_response()

        for name in call_result:
            yield self.collection(name)

    def collection(self, collection_name: str) -> "sync_collection.SyncCollection":
        if not isinstance(collection_name, str):
            raise TypeError("The collection name has to be a string.")

        return sync_collection.SyncCollection(self, collection_name)

    def drop_collection(self, collection_name: str) -> None:
        if not isinstance(collection_name, str):
            raise TypeError(
                f"The collection name has to be a string, not: {repr(collection_name)}"
            )

        call_result = native.db_drop_collection(
            self._rust_db, collection_name.encode("utf-8")
        )
        call_result.consume_into_response()

    def __getitem__(self, collection_name: str) -> "sync_collection.SyncCollection":
        return self.collection(collection_name)



from pathlib import Path
from typing import AsyncIterable, Iterable

import pytest

import notmongo as nm
from notmongo.typedefs import *

TESTS_DIRECTORY = Path(__file__).resolve().parent
TEST_DB_JSON_PATH = TESTS_DIRECTORY / "test-db.json"


def _load_test_db() -> nm.AsyncDatabase:
    return nm.AsyncDatabase.load_json(TEST_DB_JSON_PATH)


def _load_test_collection() -> nm.AsyncCollection:
    return _load_test_db().collection("animals")


async def _collect_async_generator(generator: AsyncIterable[Any]) -> List[Any]:
    result = []

    async for value in generator:
        result.append(value)

    return result


async def _verify_result_ids(
    documents: Union[nm.AsyncCollection, Iterable[BsonLike]],
    expected_ids: Iterable[BsonLike],
):
    # Get all ids
    if isinstance(documents, nm.AsyncCollection):
        documents = await _collect_async_generator(documents.find({}))

    have_ids = {doc["_id"] for doc in documents}

    expected_ids = set(expected_ids)

    # Superfluous
    superfluous_ids = have_ids - expected_ids
    assert (
        not superfluous_ids
    ), f"Superfluous result items, with these ids: `{'`, `'.join(map(str, superfluous_ids))}`"

    # Missing
    missing_ids = expected_ids - have_ids
    assert (
        not missing_ids
    ), f"Missing result items, with these ids: `{'`, `'.join(map(str, missing_ids))}`"


@pytest.mark.asyncio
async def test_count_all_documents():
    col = _load_test_collection()
    assert await col.count_all_documents() == 3


@pytest.mark.asyncio
async def test_count_documents_all():
    col = _load_test_collection()
    assert await col.count_documents({}) == 3


@pytest.mark.asyncio
async def test_count_documents_one_attriubute():
    col = _load_test_collection()
    assert await col.count_documents({"canSwim": True}) == 2


@pytest.mark.asyncio
async def test_count_documents_two_attriubutes():
    col = _load_test_collection()
    assert await col.count_documents({"canSwim": True, "legs": 4}) == 1


@pytest.mark.asyncio
async def test_delete_many_no_filter():
    col = _load_test_collection()
    await col.delete_many({})
    await _verify_result_ids(col, set())


@pytest.mark.asyncio
async def test_delete_many_with_filter():
    col = _load_test_collection()
    await col.delete_many({"legs": 4})
    await _verify_result_ids(col, {"id-1-fish"})


@pytest.mark.asyncio
async def test_delete_one_no_filter():
    col = _load_test_collection()
    await col.delete_one({})
    assert await col.count_all_documents() == 2


@pytest.mark.asyncio
async def test_find_all():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}))
    await _verify_result_ids(results, {"id-1-cat", "id-1-dog", "id-1-fish"})


@pytest.mark.asyncio
async def test_find_all_skip_1():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, skip=1))
    results = list(results)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_find_all_limit_1():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, limit=1))
    results = list(results)
    assert len(results) == 1


@pytest.mark.asyncio
async def test_find_all_with_sort():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, sort=[("age", -1)]))
    results = list(results)

    assert len(results) == 3
    assert results[0]["_id"] == "id-1-dog"
    assert results[1]["_id"] == "id-1-cat"
    assert results[2]["_id"] == "id-1-fish"


@pytest.mark.asyncio
async def test_find_all_with_sort_and_skip():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, sort=[("age", -1)], skip=2))

    assert len(results) == 1
    assert results[0]["_id"] == "id-1-fish"


@pytest.mark.asyncio
async def test_find_all_with_sort_and_limit():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, sort=[("age", -1)], limit=2))

    assert len(results) == 2
    assert results[0]["_id"] == "id-1-dog"
    assert results[1]["_id"] == "id-1-cat"


@pytest.mark.asyncio
async def test_find_all_with_sort_and_skip_and_limit():
    col = _load_test_collection()
    results = await _collect_async_generator(
        col.find({}, sort=[("age", -1)], skip=1, limit=1)
    )

    assert len(results) == 1
    assert results[0]["_id"] == "id-1-cat"


@pytest.mark.asyncio
async def test_find_all_skip_some():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, skip=2))
    assert len(results) == 1


@pytest.mark.asyncio
async def test_find_all_skip_all():
    col = _load_test_collection()
    results = await _collect_async_generator(col.find({}, skip=3))
    assert len(results) == 0


@pytest.mark.asyncio
async def test_find_one_any():
    col = _load_test_collection()
    result = await col.find_one({})

    assert result is not None


@pytest.mark.asyncio
async def test_find_one_with_filter():
    col = _load_test_collection()
    result = await col.find_one({"_id": "id-1-dog"})

    assert result is not None
    assert result["_id"] == "id-1-dog"


@pytest.mark.asyncio
async def test_find_one_with_sort():
    col = _load_test_collection()
    result = await col.find_one({}, sort=[("age", -1)])

    assert result is not None
    assert result["_id"] == "id-1-dog"


@pytest.mark.asyncio
async def test_find_one_with_sort_and_skip():
    col = _load_test_collection()
    result = await col.find_one({}, sort=[("age", -1)], skip=2)

    assert result is not None
    assert result["_id"] == "id-1-fish"


@pytest.mark.asyncio
async def test_insert_one():
    col = _load_test_collection()
    result = await col.insert_one({"_id": "id-new-document", "canSwim": True})

    assert result.inserted_id == "id-new-document"
    await _verify_result_ids(
        col, {"id-1-cat", "id-1-dog", "id-1-fish", "id-new-document"}
    )


@pytest.mark.asyncio
async def test_insert_one_without_id():
    col = _load_test_collection()

    result = await col.insert_one({"canSwim": True})

    await _verify_result_ids(
        col, {"id-1-cat", "id-1-dog", "id-1-fish", result.inserted_id}
    )


@pytest.mark.asyncio
async def test_insert_many():
    col = _load_test_collection()
    result = await col.insert_many(
        [
            {"_id": "id-new-document-1", "canSwim": True},
            {"_id": "id-new-document-2", "canSwim": True},
        ]
    )

    assert list(result.inserted_ids) == ["id-new-document-1", "id-new-document-2"]

    await _verify_result_ids(
        col,
        {"id-1-cat", "id-1-dog", "id-1-fish", "id-new-document-1", "id-new-document-2"},
    )


@pytest.mark.asyncio
async def test_insert_many_without_ids():
    col = _load_test_collection()
    result = await col.insert_many(
        [
            {"_id": "id-new-document-1", "canSwim": True},
            {"_id": "id-new-document-2", "canSwim": True},
        ]
    )
    await _verify_result_ids(
        col,
        {"id-1-cat", "id-1-dog", "id-1-fish"} | set(result.inserted_ids),
    )


@pytest.mark.asyncio
async def test_replace_one_with_match():
    col = _load_test_collection()
    await col.replace_one(
        {"_id": "id-1-fish"},
        {"_id": "id-new-document"},
        upsert=False,
    )
    await _verify_result_ids(col, {"id-1-cat", "id-1-dog", "id-new-document"})


@pytest.mark.asyncio
async def test_replace_one_without_match():
    col = _load_test_collection()
    await col.replace_one(
        {"_id": "id-nonexistent"},
        {"_id": "id-new-document"},
        upsert=False,
    )
    await _verify_result_ids(col, {"id-1-cat", "id-1-dog", "id-1-fish"})


@pytest.mark.asyncio
async def test_replace_one_without_match_but_upsert():
    col = _load_test_collection()
    await col.replace_one(
        {"_id": "id-nonexistent"},
        {"_id": "id-new-document"},
        upsert=True,
    )
    await _verify_result_ids(
        col, {"id-1-cat", "id-1-dog", "id-1-fish", "id-new-document"}
    )


@pytest.mark.asyncio
async def test_update_one_with_match():
    col = _load_test_collection()
    await col.update_one(
        {"_id": "id-1-fish"},
        {"$set": {"_id": "id-updated-document"}},
        upsert=False,
    )
    await _verify_result_ids(col, {"id-1-cat", "id-1-dog", "id-updated-document"})


@pytest.mark.asyncio
async def test_update_one_without_match():
    col = _load_test_collection()
    await col.update_one(
        {"_id": "id-nonexistent"},
        {"$set": {"_id": "id-updated-document"}},
        upsert=False,
    )
    await _verify_result_ids(col, {"id-1-cat", "id-1-dog", "id-1-fish"})


@pytest.mark.asyncio
async def test_update_one_without_match_but_upsert():
    col = _load_test_collection()
    await col.update_one(
        {"_id": "id-nonexistent"},
        {"$set": {"_id": "id-upserted-document"}},
        upsert=True,
    )
    await _verify_result_ids(
        col, {"id-1-cat", "id-1-dog", "id-1-fish", "id-upserted-document"}
    )


@pytest.mark.asyncio
async def test_update_many_with_match():
    col = _load_test_collection()
    update_result = await col.update_many(
        {"legs": 4},
        {"$set": {"legs": "updated"}, "$unset": "canSwim"},
        upsert=False,
    )

    assert update_result.matched_count == 2
    assert update_result.modified_count == 2
    assert update_result.upserted_id is None

    async for result in col.find():
        if result["_id"] == "id-1-fish":
            assert "canSwim" in result
        else:
            assert result["legs"] == "updated"
            assert "canSwim" not in result


@pytest.mark.asyncio
async def test_update_many_unset_nonexistent_field():
    col = _load_test_collection()
    update_result = await col.update_many(
        {"legs": 4},
        {"$unset": "noSuchField"},
        upsert=False,
    )

    assert update_result.matched_count == 2
    assert update_result.modified_count == 2
    assert update_result.upserted_id is None


@pytest.mark.asyncio
async def test_update_many_without_match():
    col = _load_test_collection()
    update_result = await col.update_many(
        {"legs": 5},
        {"$set": {"legs": "updated"}},
        upsert=False,
    )

    assert update_result.matched_count == 0
    assert update_result.modified_count == 0
    assert update_result.upserted_id is None

    async for result in col.find():
        if result["_id"] == "id-1-fish":
            assert "canSwim" in result
        else:
            assert result["legs"] == 4
            assert "canSwim" in result


@pytest.mark.asyncio
async def test_update_many_without_match_but_upsert():
    col = _load_test_collection()
    update_result = await col.update_many(
        {"_id": "id-nonexistent"},
        {"$set": {"_id": "id-upserted-document"}},
        upsert=True,
    )

    assert update_result.matched_count == 0
    assert update_result.modified_count == 1
    assert update_result.upserted_id is not None

    await _verify_result_ids(
        col, {"id-1-cat", "id-1-dog", "id-1-fish", "id-upserted-document"}
    )

    async for result in col.find():
        if result["_id"] == "id-1-fish":
            assert "canSwim" in result
        elif result["_id"] == "id-upserted-document":
            assert len(result) == 1
        else:
            assert result["legs"] == 4
            assert "canSwim" in result



from typing import List, Optional

import bson
from bson import BSON

from .typedefs import *


class DeleteResult:
    def __init__(self, n_deleted: int):
        self.n_deleted = n_deleted

    @classmethod
    def decode(cls, raw: BsonLike) -> "DeleteResult":
        return cls(raw["nDeleted"])


class UpdateResult:
    def __init__(
        self,
        matched_count: int,
        modified_count: int,
        upserted_id: Optional[BSON],
    ):
        self.matched_count = matched_count
        self.modified_count = modified_count
        self.upserted_id = upserted_id

    @classmethod
    def decode(cls, raw: BSON) -> "UpdateResult":
        return cls(
            raw["matchedCount"],
            raw["modifiedCount"],
            raw["upsertedId"],
        )


class InsertOneResult:
    def __init__(
        self,
        inserted_id: BSON,
    ):
        self.inserted_id = inserted_id

    @classmethod
    def decode(cls, raw: BSON) -> "InsertOneResult":
        return cls(
            raw["insertedId"],
        )


class InsertManyResult:
    def __init__(
        self,
        inserted_ids: List[BSON],
    ):
        self.inserted_ids = inserted_ids

    @classmethod
    def decode(cls, raw: BSON) -> "InsertManyResult":
        return cls(
            raw["insertedIds"],
        )



from typing import Any, Dict, Iterable, List, Optional, Union

from . import native, results, sync_database
from .typedefs import *


def _unsupported(name: str, value: Any, default: Any):
    if value != default:
        print(f'Warning: The "{name}" parameter is not supported and will be ignored.')


class SyncCollection:
    def __init__(self, _db: "sync_database.SyncDatabase", name: str):
        self._db = _db
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def _notify_persistence_of_change(
        self,
        result: Union[
            results.DeleteResult,
            results.UpdateResult,
            results.InsertOneResult,
            results.InsertManyResult,
        ],
    ):
        # Check if a change was actually made
        if isinstance(result, results.DeleteResult) and result.n_deleted == 0:
            return

        if isinstance(result, results.UpdateResult) and result.modified_count == 0:
            return

        if isinstance(result, results.InsertOneResult) and result.inserted_id is None:
            return

        if isinstance(result, results.InsertManyResult) and not result.inserted_ids:
            return

        # Something has changed, notify the persistence
        if self._db._persistence is not None:
            self._db._persistence.notify_change()

    def count_all_documents(
        self,
    ) -> int:
        call_result = native.collection_count_all_documents(
            self._db._rust_db,
            self._name.encode("utf-8"),
        )
        return native.wrapped_bson.consume_into_response(call_result)  # type: ignore

    def count_documents(
        self,
        filter: BsonLike,  # Default value?
        session=None,
        skip: int = 0,
        limit: int = 0,
        maxTimeMs: Optional[int] = None,
        collation=None,
        hint=None,
    ) -> int:
        _unsupported("session", session, None)
        _unsupported("maxTimeMs", maxTimeMs, None)
        _unsupported("collation", collation, None)
        _unsupported("hint", hint, None)

        call_result = native.collection_count_documents(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
            skip,
            limit,
        )
        return native.wrapped_bson.consume_into_response(call_result)  # type: ignore

    def delete_many(
        self, filter: BsonLike, collation=None, hint=None, session=None
    ) -> results.DeleteResult:
        _unsupported("session", session, None)
        _unsupported("collation", collation, None)
        _unsupported("hint", hint, None)

        call_result = native.collection_delete_many(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.DeleteResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result

    def delete_one(
        self, filter: BsonLike, collation=None, hint=None, session=None
    ) -> results.DeleteResult:
        _unsupported("session", session, None)
        _unsupported("collation", collation, None)
        _unsupported("hint", hint, None)

        call_result = native.collection_delete_one(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.DeleteResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result

    def drop(self):
        self._db.drop_collection(self._name)

    def find(
        self,
        filter: BsonLike = {},
        projection: Union[List[str], Dict[str, bool]] = {},
        skip: int = 0,
        limit: int = 0,
        no_cursor_timeout: bool = False,
        cursor_type: str = "non_tailable",
        sort: SortList = [],
        allow_partial_results: bool = False,
        oplog_replay: bool = False,
        batch_size: Optional[int] = None,
        manipulate: Optional[bool] = None,
        collation=None,
        show_record_id: bool = False,
        snapshot: bool = False,
        hint=None,
        max_time_ms: Optional[int] = None,
        max_scan: Optional[int] = None,
        min=None,
        max=None,
        comment: Optional[str] = None,
        session=None,
        modifiers: dict = {},
        allow_disk_use: bool = False,
    ) -> Iterable[BsonLike]:
        _unsupported("no_cursor_timeout", no_cursor_timeout, False)
        _unsupported("cursor_type", cursor_type, "non_tailable")
        _unsupported("allow_partial_results", allow_partial_results, False)
        _unsupported("oplog_replay", oplog_replay, False)
        _unsupported("batch_size", batch_size, None)
        _unsupported("manipulate", manipulate, None)
        _unsupported("collation", collation, None)
        _unsupported("show_record_id", show_record_id, False)
        _unsupported("snapshot", snapshot, False)
        _unsupported("hint", hint, None)
        _unsupported("max_time_ms", max_time_ms, None)
        _unsupported("max_scan", max_scan, None)
        _unsupported("min", min, None)
        _unsupported("max", max, None)
        _unsupported("comment", comment, None)
        _unsupported("session", session, None)
        _unsupported("modifiers", modifiers, {})
        _unsupported("allow_disk_use", allow_disk_use, False)

        call_result = native.collection_find(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
            native.wrapped_bson.from_bson({"projection": projection}),
            skip,
            limit,
            native.wrapped_bson.from_bson({"sort": sort}),
        )
        return native.wrapped_bson.consume_into_response(call_result)  # type: ignore

    def find_one(
        self,
        filter: BsonLike = {},
        projection: Union[List[str], Dict[str, bool]] = {},
        skip: int = 0,
        limit: int = 0,
        no_cursor_timeout: bool = False,
        cursor_type: str = "non_tailable",
        sort: SortList = [],
        allow_partial_results: bool = False,
        oplog_replay: bool = False,
        batch_size: Optional[int] = None,
        manipulate: Optional[bool] = None,
        collation=None,
        show_record_id: bool = False,
        snapshot: bool = False,
        hint=None,
        max_time_ms: Optional[int] = None,
        max_scan: Optional[int] = None,
        min=None,
        max=None,
        comment: Optional[str] = None,
        session=None,
        modifiers: dict = {},
        allow_disk_use: bool = False,
    ) -> Optional[BsonLike]:
        if limit not in (0, 1):
            print(f"Warning: Using `find_one` with a limit of `{limit}`!")

        _unsupported("no_cursor_timeout", no_cursor_timeout, False)
        _unsupported("cursor_type", cursor_type, "non_tailable")
        _unsupported("allow_partial_results", allow_partial_results, False)
        _unsupported("oplog_replay", oplog_replay, False)
        _unsupported("batch_size", batch_size, None)
        _unsupported("manipulate", manipulate, None)
        _unsupported("collation", collation, None)
        _unsupported("show_record_id", show_record_id, False)
        _unsupported("snapshot", snapshot, False)
        _unsupported("hint", hint, None)
        _unsupported("max_time_ms", max_time_ms, None)
        _unsupported("max_scan", max_scan, None)
        _unsupported("min", min, None)
        _unsupported("max", max, None)
        _unsupported("comment", comment, None)
        _unsupported("session", session, None)
        _unsupported("modifiers", modifiers, {})
        _unsupported("allow_disk_use", allow_disk_use, False)

        call_result = native.collection_find_one(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
            native.wrapped_bson.from_bson({"projection": projection}),
            skip,
            native.wrapped_bson.from_bson({"sort": sort}),
        )
        return native.wrapped_bson.consume_into_response(call_result)

    def insert_one(
        self,
        document: BsonLike,
        bypass_document_validation: bool = False,
        session: bool = None,
    ) -> results.InsertOneResult:
        _unsupported("bypass_document_validation", bypass_document_validation, False)
        _unsupported("session", session, None)

        call_result = native.collection_insert_one(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(document),
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.InsertOneResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result

    def insert_many(
        self,
        documents: Iterable[BsonLike],
        ordered: bool = True,
        bypass_document_validation: bool = False,
        session: bool = None,
    ) -> results.InsertManyResult:
        # "ordered" is ignored, but insertions in Not Mongo are always ordered
        # anyways.
        _unsupported("bypass_document_validation", bypass_document_validation, False)
        _unsupported("session", session, None)

        call_result = native.collection_insert_many(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson({"documents": list(documents)}),
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.InsertManyResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result

    def replace_one(
        self,
        filter: BsonLike,
        replacement: BsonLike,
        upsert=False,
        bypass_document_validation=False,
        collation=None,
        hint=None,
        session=None,
    ) -> results.UpdateResult:
        _unsupported("bypass_document_validation", bypass_document_validation, False)
        _unsupported("collaction", collation, None)
        _unsupported("hint", hint, None)
        _unsupported("session", session, None)

        call_result = native.collection_replace_one(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
            native.wrapped_bson.from_bson(replacement),
            upsert,
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.UpdateResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result

    def update_one(
        self,
        filter: BsonLike,
        update: BsonLike,
        upsert=False,
        bypass_document_validation=False,
        collation=None,
        array_filters=None,
        hint=None,
        session=None,
    ) -> results.UpdateResult:
        _unsupported("bypass_document_validation", bypass_document_validation, False)
        _unsupported("collaction", collation, None)
        _unsupported("array_filters", array_filters, None)
        _unsupported("hint", hint, None)
        _unsupported("session", session, None)

        call_result = native.collection_update_one(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
            native.wrapped_bson.from_bson(update),
            upsert,
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.UpdateResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result

    def update_many(
        self,
        filter: BsonLike,
        update: BsonLike,
        upsert=False,
        bypass_document_validation=False,
        collation=None,
        array_filters=None,
        hint=None,
        session=None,
    ) -> results.UpdateResult:
        _unsupported("bypass_document_validation", bypass_document_validation, False)
        _unsupported("collation", collation, None)
        _unsupported("array_filters", array_filters, None)
        _unsupported("hint", hint, None)
        _unsupported("session", session, None)

        call_result = native.collection_update_many(
            self._db._rust_db,
            self._name.encode("utf-8"),
            native.wrapped_bson.from_bson(filter),
            native.wrapped_bson.from_bson(update),
            upsert,
        )

        call_result = native.wrapped_bson.consume_into_response(call_result)
        call_result = results.UpdateResult.decode(call_result)
        self._notify_persistence_of_change(call_result)
        return call_result



import asyncio
from typing import AsyncIterable, Dict, Iterable, List, Optional, Union

from . import results, sync_collection
from .typedefs import *


def _make_wrapped(method):
    async def async_method(self, *args, **kwargs):
        sync_method = getattr(self._collection, method.__name__)

        async with self._db_lock:
            return await asyncio.to_thread(sync_method, *args, **kwargs)

    return async_method


class AsyncCollection:
    """
    Asynchronous wrapper for SyncCollection. While this won't provide any of the
    advantages of actual async queries, this allows code to be written
    asynchronously, and then easily switched to the real MongoDB.
    """

    def __init__(
        self,
        collection: sync_collection.SyncCollection,
        db_lock: asyncio.Lock,
    ):
        self._collection = collection
        self._db_lock = db_lock

    @property
    def name(self) -> str:
        return self._collection.name

    @_make_wrapped
    async def count_all_documents(
        self,
    ) -> int:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def count_documents(
        self,
        filter: BsonLike,  # Default value?
        session=None,
        skip: int = 0,
        limit: int = 0,
        maxTimeMs: Optional[int] = None,
        collation=None,
        hint=None,
    ) -> int:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def delete_many(
        self, filter: BsonLike, collation=None, hint=None, session=None
    ) -> results.DeleteResult:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def delete_one(
        self, filter: BsonLike, collation=None, hint=None, session=None
    ) -> results.DeleteResult:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def drop(self):
        pass

    async def find(
        self,
        filter: BsonLike = {},
        projection: Union[List[str], Dict[str, bool]] = {},
        skip: int = 0,
        limit: int = 0,
        no_cursor_timeout: bool = False,
        cursor_type: str = "non_tailable",
        sort: SortList = [],
        allow_partial_results: bool = False,
        oplog_replay: bool = False,
        batch_size: Optional[int] = None,
        manipulate: Optional[bool] = None,
        collation=None,
        show_record_id: bool = False,
        snapshot: bool = False,
        hint=None,
        max_time_ms: Optional[int] = None,
        max_scan: Optional[int] = None,
        min=None,
        max=None,
        comment: Optional[str] = None,
        session=None,
        modifiers: dict = {},
        allow_disk_use: bool = False,
    ) -> AsyncIterable[BsonLike]:
        result_iter = self._collection.find(
            filter=filter,
            projection=projection,
            skip=skip,
            limit=limit,
            no_cursor_timeout=no_cursor_timeout,
            cursor_type=cursor_type,
            sort=sort,
            allow_partial_results=allow_partial_results,
            oplog_replay=oplog_replay,
            batch_size=batch_size,
            manipulate=manipulate,
            collation=collation,
            show_record_id=show_record_id,
            snapshot=snapshot,
            hint=hint,
            max_time_ms=max_time_ms,
            max_scan=max_scan,
            min=min,
            max=max,
            comment=comment,
            session=session,
            modifiers=modifiers,
            allow_disk_use=allow_disk_use,
        )

        for result in result_iter:
            yield result

    @_make_wrapped
    async def find_one(
        self,
        filter: BsonLike = {},
        projection: Union[List[str], Dict[str, bool]] = {},
        skip: int = 0,
        limit: int = 0,
        no_cursor_timeout: bool = False,
        cursor_type: str = "non_tailable",
        sort: SortList = [],
        allow_partial_results: bool = False,
        oplog_replay: bool = False,
        batch_size: Optional[int] = None,
        manipulate: Optional[bool] = None,
        collation=None,
        show_record_id: bool = False,
        snapshot: bool = False,
        hint=None,
        max_time_ms: Optional[int] = None,
        max_scan: Optional[int] = None,
        min=None,
        max=None,
        comment: Optional[str] = None,
        session=None,
        modifiers: dict = {},
        allow_disk_use: bool = False,
    ) -> Optional[BsonLike]:
        pass

    @_make_wrapped
    async def insert_one(
        self,
        document: BsonLike,
        bypass_document_validation: bool = False,
        session: bool = None,
    ) -> results.InsertOneResult:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def insert_many(
        self,
        documents: Iterable[BsonLike],
        ordered: bool = True,
        bypass_document_validation: bool = False,
        session: bool = None,
    ) -> results.InsertManyResult:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def replace_one(
        self,
        filter: BsonLike,
        replacement: BsonLike,
        upsert=False,
        bypass_document_validation=False,
        collation=None,
        hint=None,
        session=None,
    ) -> results.UpdateResult:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def update_one(
        self,
        filter: BsonLike,
        update: BsonLike,
        upsert=False,
        bypass_document_validation=False,
        collation=None,
        array_filters=None,
        hint=None,
        session=None,
    ) -> results.UpdateResult:
        raise RuntimeError("Unreachable")

    @_make_wrapped
    async def update_many(
        self,
        filter: BsonLike,
        update: BsonLike,
        upsert=False,
        bypass_document_validation=False,
        collation=None,
        array_filters=None,
        hint=None,
        session=None,
    ) -> results.UpdateResult:
        raise RuntimeError("Unreachable")



import distutils.command.build
import shutil
import tarfile
from pathlib import Path
import urllib.request

from setuptools import setup

from setuptools_rust import Binding, RustExtension
import notmongo_meta as meta


PROJECT_ROOT_DIR = Path(__file__).resolve().parent
SRC_DIR = PROJECT_ROOT_DIR / "src"
BUILD_DIR = PROJECT_ROOT_DIR / "build"

CRATE_TARBALL_PATH = SRC_DIR / "notmongo-rs.tar.gz"
RUST_SOURCE_DIR = SRC_DIR / "notmongo-rs"
PY_SOURCE_DIR = SRC_DIR / "notmongo"


# Download notmongo-rs from crates.io
#
# TODO: This should really be done in the build function, however `setup` fails
#       if the rust sources don't already exist.
def ensure_rust_source_code_exists():
    # Fetch the tarball, if needed
    if not CRATE_TARBALL_PATH.exists():
        print('Downloading "notmongo-rs" from crates.io')
        download_url = (
            f"https://crates.io/api/v1/crates/notmongo/{meta.__version__}/download"
        )

        with urllib.request.urlopen(download_url) as response:
            with CRATE_TARBALL_PATH.open("wb") as outfile:
                shutil.copyfileobj(response, outfile)

    # Unzip the archive, if needed
    if not RUST_SOURCE_DIR.exists():
        print("Extracting rust source")
        with tarfile.open(CRATE_TARBALL_PATH) as zfile:
            zfile.extractall(SRC_DIR)

        shutil.move(SRC_DIR / f"notmongo-{meta.__version__}", RUST_SOURCE_DIR)

ensure_rust_source_code_exists()


class BuildCommand(distutils.command.build.build):
    def run(self):
        # TODO: Would be great to download the rust source here, but this causes
        # `setup` to fail

        # Chain to the regular build function
        super().run()

        # Find the built shared object as well as the built `notmongo` module
        # and move the shared object into the module
        so_path = next(iter(BUILD_DIR.glob("**/notmongo-rs*.so")))
        python_module_path = so_path.parent / "notmongo"

        assert python_module_path.exists(), f"Could not find python module at {python_module_path}"

        shared_objects_dir = python_module_path / "shared_objects"
        shared_objects_dir.mkdir(exist_ok=True)
        shutil.move(so_path, shared_objects_dir / "libnotmongo.so")

        shutil.copy(PROJECT_ROOT_DIR / 'notmongo_meta.py', python_module_path / 'meta.py')


setup(
    name="notmongo",
    version=meta.__version__,
    rust_extensions=[
        RustExtension(
            "notmongo-rs",
            binding=Binding.NoBinding,
            path="src/notmongo-rs/Cargo.toml",
        )
    ],
    packages=["notmongo-rs", "notmongo"],
    package_dir={"": "src"},
    zip_safe=False,
    cmdclass={"build": BuildCommand},
    extras_require={
        "dev": [
            "black",
            "isort",
            "pre_commit",
            "pytest",
            "pytest-asyncio",
        ]
    },
)



import time
from pathlib import Path
from typing import Literal, Optional, Union

from . import sync_database
from .typedefs import *


class BasePersistence:
    def __init__(self):
        self.db: Optional[
            "sync_database.SyncDatabase"
        ] = None  # This is set by the database when the persistence is assigned to it

    def save_now(self):
        if self.db is None:
            raise ValueError("Assign the persistence to a database before using it")

    def notify_change(self):
        pass


# TODO: Saves are only triggered on changes to the database. This means that if
# one operation is made, and then a long pause the changes won't be written to
# disk until another change is performed.
class FilePersistence(BasePersistence):
    def __init__(
        self,
        file_path: Union[Path, str],
        *,
        n_backups: int = 3,
        file_format: Literal["json", "bson"] = "json",
        save_interval: float = 60,
        use_relaxed_extjson: bool = False,
        format_json: bool = False,
        compression: Compression = None,
    ):
        super().__init__()

        if isinstance(file_path, str):
            file_path = Path(file_path)

        if not isinstance(file_path, Path):
            raise TypeError(file_path)

        if not isinstance(n_backups, int):
            raise TypeError(n_backups)

        if n_backups < 0:
            raise ValueError("The number of backups cannot be negative.")

        if file_format == "bson" and compression is not None:
            raise ValueError("Compression is not currently supported with BSON files")

        file_format = file_format.lower()  # type: ignore
        if file_format not in ("json", "bson"):
            raise ValueError('The file format needs to be either "json" or "bson"')

        if not isinstance(format_json, bool):
            raise TypeError(format_json)

        self.file_path = file_path
        self.n_backups = n_backups
        self.file_format = file_format
        self.save_interval = save_interval
        self.use_relaxed_extjson = use_relaxed_extjson
        self.format_json = format_json
        self.compression: Compression = compression

        self.is_dirty = False
        self.last_save_time = 0

    def save_now(self):
        # Is the database even set?
        if self.db is None:
            raise ValueError("Assign the persistence to a database before using it")

        # Save the database
        if self.file_format == "json":
            self.db.dump_json(
                self.file_path,
                relaxed=self.use_relaxed_extjson,
                format=self.format_json,
                n_backups=self.n_backups,
                compression=self.compression,
            )
        else:
            self.db.dump_bson(
                self.file_path,
                n_backups=self.n_backups,
                compression=None,
            )

        # Keep track of this save
        self.is_dirty = False
        self.last_save_time = time.time()

    def notify_change(self):
        self.is_dirty = True

        if (
            self.save_interval == 0
            or self.last_save_time + self.save_interval < time.time()
        ):
            self.save_now()

    def __del__(self):
        if self.is_dirty:
            self.save_now()



from .async_collection import *
from .async_database import *
from .persistence import *
from .sync_collection import *
from .sync_database import *
from .typedefs import *
from .meta import __version__
